/* Copyright 2012 Mozilla Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import { FindState } from "./pdf_find_controller.js";

const MATCHES_COUNT_LIMIT = 1000;

/**
 * Creates a "search bar" given a set of DOM elements that act as controls
 * for searching or for setting search preferences in the UI. This object
 * also sets up the appropriate events for the controls. Actual searching
 * is done by PDFFindController.
 */
class PDFFindBar {
  constructor(options, eventBus) {
    this.opened = false;

    this.bar = options.bar;
    this.toggleButton = options.toggleButton;
    this.findField = options.findField;
    this.emptyInput = options.emptyInput;
    this.highlightAll = options.highlightAllCheckbox || { checked: true };
    this.caseSensitive = options.caseSensitiveCheckbox || { checked: false };
    this.matchDiacritics = options.matchDiacriticsCheckbox || { checked: false };
    this.entireWord = options.entireWordCheckbox || { checked: false };
    this.findMsg = options.findMsg;
    this.findResultsCount = options.findResultsCount;
    this.findPreviousButton = options.findPreviousButton;
    this.findNextButton = options.findNextButton;
    this.eventBus = eventBus;
    this.l10n = null;
    this.findbarView = options.findbarView;
    
    // Add event listeners to the DOM elements.
    this.findField.addEventListener("input", () => {
      this.dispatchEvent("");
      this.emptyInput.classList.remove("hidden");
      if (!this.findField.value) {
        this.emptyInput.classList.add("hidden");
      }
    });

    this.bar.addEventListener("keydown", e => {
      switch (e.keyCode) {
        case 13: // Enter
          if (e.target === this.findField) {
            this.dispatchEvent("again", e.shiftKey);
          }
          break;
        case 27: // Escape
          this.close();
          break;
      }
    });

    this.findPreviousButton.addEventListener("click", () => {
      this.dispatchEvent("again", true);
    });

    this.findNextButton.addEventListener("click", () => {
      this.dispatchEvent("again", false);
    });

    this.eventBus._on("resize", this.#adjustWidth.bind(this));

    this.emptyInput.addEventListener("click", () => {
      this.empty();
    });
  }

  reset() {
    this.updateUIState();
  }

  dispatchEvent(type, findPrev = false, value = null) {
    this.eventBus.dispatch("find", {
      source: this,
      type,
      query: value || this.findField.value,
      phraseSearch: true,
      caseSensitive: this.caseSensitive.checked,
      entireWord: this.entireWord.checked,
      highlightAll: this.highlightAll.checked,
      findPrevious: findPrev,
      matchDiacritics: this.matchDiacritics.checked,
    });
  }

  updateUIState(state, previous, matchesCount) {
    let status = "";
    let close = false

    switch (state) {
      case FindState.FOUND:
        break;
      case FindState.PENDING:
        status = "pending";
        break;
      case FindState.NOT_FOUND:
        status = "notFound";
        break;
      case FindState.WRAPPED:
        break;
      case FindState.CLOSE:
        close = true
        break;
    }
    this.findField.setAttribute("data-status", status);
    this.findField.setAttribute("aria-invalid", state === FindState.NOT_FOUND);

    this.#adjustWidth();

    this.updateResultsCount(matchesCount, close);
  }

  updateResultsCount({ current = 0, total = 0 } = {}, close = false) {
    let matchCountMsg = '';
    if  (close) {
      matchCountMsg = ''
      this.findPreviousButton.classList.add("disabled")
      this.findNextButton.classList.add("disabled")
    }else if (total > 0) {
      matchCountMsg = current + '/' + total + ' results found'
      this.findPreviousButton.classList.remove("disabled")
      this.findNextButton.classList.remove("disabled")
    } else {
      matchCountMsg = '0/0 results found'
      this.findPreviousButton.classList.add("disabled")
      this.findNextButton.classList.add("disabled")
      this.eventBus.dispatch('updateFindMatches', [])
    }
    this.findResultsCount.textContent = matchCountMsg;
    // Since `updateResultsCount` may be called from `PDFFindController`,
    // ensure that the width of the findbar is always updated correctly.
    this.#adjustWidth();
  }

  open() {
    if (!this.opened) {
      this.opened = true;
      this.toggleButton.classList.add("toggled");
      this.toggleButton.setAttribute("aria-expanded", "true");
      // this.bar.classList.remove("hidden");
    }
    this.findField.select();
    this.findField.focus();

    this.#adjustWidth();
  }

  close() {
    if (!this.opened) {
      return;
    }
    this.opened = false;
    this.toggleButton.classList.remove("toggled");
    this.toggleButton.setAttribute("aria-expanded", "false");
    // this.bar.classList.add("hidden");
    this.empty();
    this.findResultsCount.textContent = ''
    this.findPreviousButton.classList.add("disabled")
    this.findNextButton.classList.add("disabled")
    this.eventBus.dispatch("findbarclose", { source: this });
  }

  toggle() {
    if (this.opened) {
      this.close();
    } else {
      this.open();
    }
  }

  #adjustWidth() {
    if (!this.opened) {
      return;
    }

    // The find bar has an absolute position and thus the browser extends
    // its width to the maximum possible width once the find bar does not fit
    // entirely within the window anymore (and its elements are automatically
    // wrapped). Here we detect and fix that.
    this.bar.classList.remove("wrapContainers");

    const findbarHeight = this.bar.clientHeight;
    const inputContainerHeight = this.bar.firstElementChild.clientHeight;

    if (findbarHeight > inputContainerHeight) {
      // The findbar is taller than the input container, which means that
      // the browser wrapped some of the elements. For a consistent look,
      // wrap all of them to adjust the width of the find bar.
      this.bar.classList.add("wrapContainers");
    }
  }

  empty() {
    this.findField.value = '';
    const inputEvent = new Event("input");
    this.findField.dispatchEvent(inputEvent);
    this.eventBus.dispatch('updateFindMatches', [])
    this.emptyInput.classList.add("hidden");
  }
}

export { PDFFindBar };
